import time
from typing import Literal, Optional, Union

import cloudscraper
from bs4 import BeautifulSoup
from bs4.element import Tag

SCRAPER = cloudscraper.create_scraper(browser={"browser": "chrome", "platform": "android", "desktop": False})

# check if subtitle is hearing impaired or not
def is_sub_hi(a1: Tag) -> Optional[bool]:
    """
    Check if subtitle is hearing impaired or not

    Args:
        a1 (Tag): lxml tag

    Returns:
        str: True if str is a40, False if str is a41
    """
    a1_parent = a1.parent
    a40 = a1_parent.find("td", class_="a40")  # non-hearing impaired
    a41 = a1_parent.find("td", class_="a41")  # hearing imparted
    if a40 is None:
        return True
    elif a41 is None:
        return False

    return None


# search for title
def search_for_title(url: str) -> (Union[dict[str, str], Literal["ERROR: CAPTCHA PROTECTION"]]):
    """
    Search subscene for matching titles

    Args:
        url (str): search url

    Returns:
        dict | str: title, url if found, None if not found
    """
    titles: dict[str, str] = {}
    source = SCRAPER.get(url)
    scontent = source.content
    doc = BeautifulSoup(scontent, "lxml")
    doc_result = doc.find("div", class_="search-result")
    if doc_result is None:
        doc_captcha = doc.find("h2", text="Why do I have to complete a CAPTCHA?")
        if doc_captcha.text == "Why do I have to complete a CAPTCHA?":
            return "ERROR: CAPTCHA PROTECTION"
    results = doc_result.find_all("div", class_="title")
    for i in results:
        get_title = i.get_text().strip().replace("'", "").replace(":", "").lower()
        link = [a["href"] for a in i.find_all("a", href=True) if a.text]
        titles[get_title] = f"https://subscene.com/{link[0]}"
    return titles


# search title(s) for subtitle
def search_title_for_sub(language: str, hearing_impaired: Union[bool, str], url: str) -> dict[str, str]:
    """
    Search for subtitles matching the provided titles

    Args:
        language (str): subtitle language
        hearing_impaired (str): if the user wants hearing impaired subtitles, regular or both
        url (str):

    Returns:
        dict: title, url, percentage match
    """
    searching = True
    subtitles: dict[str, str] = {}
    while searching:
        source = SCRAPER.get(url)
        scontent = source.content
        doc = BeautifulSoup(scontent, "lxml")
        tbody = doc.find("tbody")
        if tbody is not None:
            tda1 = tbody.find_all("td", class_="a1")
            searching = False
        else:
            time.sleep(1)

    for a1 in tda1:
        sub_hi = is_sub_hi(a1)
        if hearing_impaired != "Both" and hearing_impaired != sub_hi:
            continue
        if language == a1.span.get_text().strip():
            _release_name = [x.get_text().strip() for x in a1.find("a")]
            release_name = _release_name[-2]
            if " " in release_name:
                release_name = release_name.replace(" ", ".")
            link = [y["href"] for y in a1.find_all("a", href=True) if y.text]
            subtitles[release_name] = f"https://subscene.com/{link[0]}"

    return subtitles


# get download url for subtitle(s)
def get_download_url(url: str) -> str:
    """
    Get the download url for the subtitle

    Args:
        url (str): url for the subtitle

    Returns:
        dict: name of subtitle, url for zip file
    """
    source = SCRAPER.get(url).text
    doc = BeautifulSoup(source, "lxml")
    _link = [dl["href"] for dl in doc.find_all("a", href=True, id="downloadButton")]
    download_url = f"https://subscene.com/{_link[0]}"
    return download_url



import os
import tkinter as tk
from tkinter import ttk

import sv_ttk

from subsearch.data import __video_directory__
from subsearch.gui import tkdata, widget_root
from subsearch.scraper import subscene_soup
from subsearch.utils import file_manager

TKWINDOW = tkdata.Window()
TKCOLOR = tkdata.Color()
TKFONT = tkdata.Font()
# file with subtitles and corresponding dl links
def read_tmp_file():
    file = os.path.join(__video_directory__, "__subsearch__dl_data.tmp")
    with open(file, "r") as f:
        return [line.strip() for line in f]


# download said subtitle to the folder with the video file in it
class DownloadList(tk.Frame):
    def __init__(self, parent):
        tk.Frame.__init__(self, parent)
        # listbox for the subtitles
        self.extent = 0
        self.sublist = read_tmp_file()
        self.hs = ttk.Scrollbar(root, orient="vertical", style="Vertical.TScrollbar")
        sub_listbox = tk.Listbox(
            root,
            bg=TKCOLOR.dark_grey,
            fg=TKCOLOR.light_grey,
            font=TKFONT.cas8b,
            bd=0,
            border=0,
            borderwidth=0,
            highlightthickness=0,
            yscrollcommand=self.hs.set,
        )
        sub_listbox.place(
            height=TKWINDOW.height - 60,
            width=TKWINDOW.width - 20,
            relx=0.5,
            rely=0.525,
            bordermode="inside",
            anchor="center",
        )
        self.sub_listbox = sub_listbox
        self.fill_listbox()
        # custom scrollbar
        scrollbar_lengt = 50
        self.scrollbar_lengt = scrollbar_lengt
        self.scrollbar_lenght_half = round(scrollbar_lengt / 2)
        style = ttk.Style()
        style.theme_use("sun-valley-dark")
        style.element_options("Vertical.TScrollbar.thumb")
        # configure the style
        style.configure(
            "Vertical.TScrollbar",
            gripcount=0,
            state="disable",
            relief="flat",
            borderwidth=0,
            bd=0,
            arrowsize=24,
        )

        self.hs.place(x=TKWINDOW.width - 28, y=51, bordermode="inside", height=633)
        self.hs.config(command=self.sub_listbox.yview)
        self.hs.lift()

    def fill_listbox(self):
        dicts_names = {}
        dicts_urls = {}
        # fil list box with all available subtitles that were found and not downloaded
        for enu, item in enumerate(self.sublist):
            x = item.split(" ")
            match = f"{x[0]} {x[1]}"
            name = x[2]
            url = x[-1]
            txt = f"{match} {name}"
            self.sub_listbox.insert(tk.END, f"{match} {name}\n")
            self.sub_listbox.bind("<ButtonPress-1>", self.mouse_b1_press)
            dicts_names[enu] = txt
            dicts_urls[enu] = url
            self.dicts_urls = dicts_urls
            self.dicts_names = dicts_names

    def mouse_b1_press(self, event):
        self.sub_listbox.bind("<<ListboxSelect>>", self.download_button)

    def mouse_b1_release(self, event):
        self.sub_listbox.bind("<ButtonPress-1>", self.mouse_b1_press)

    def download_button(self, event):
        self.sub_listbox.unbind("<<ListboxSelect>>")
        self.sub_listbox.bind("<ButtonRelease-1>", self.mouse_b1_release)
        _i = str(self.sub_listbox.curselection())
        _i = _i.replace("(", "")
        _i = _i.replace(")", "")
        items = _i.replace(",", "")
        _error = False
        for (number, url), (name) in zip(self.dicts_urls.items(), self.dicts_names.values()):
            if number == int(items):
                self.sub_listbox.delete(int(number))
                self.sub_listbox.insert(int(number), f"»»» DOWNLOADING «««")
                self.sub_listbox.itemconfig(int(number), {"fg": TKCOLOR.blue})
                try:
                    dl_url = subscene_soup.get_download_url(url)
                    _name = name.replace("/", "").replace("\\", "").split(": ")
                    path = f"{__video_directory__}\\__subsearch__{_name[-1]}.zip"
                    item = path, dl_url, 1, 1
                    file_manager.download_zip(item)
                    if _error is False:
                        file_manager.extract_files(__video_directory__, ".zip")
                        file_manager.clean_up(__video_directory__, ".zip")
                        file_manager.clean_up(__video_directory__, ").nfo")
                        self.sub_listbox.delete(int(number))
                        self.sub_listbox.insert(int(number), f"✔ {name}")
                        self.sub_listbox.itemconfig(int(number), {"fg": TKCOLOR.green})
                        _error = False
                except OSError:
                    _error = True
                    self.sub_listbox.delete(int(number))
                    self.sub_listbox.insert(int(number), f"⚠⚠⚠ Download failed ⚠⚠⚠")
                    self.sub_listbox.itemconfig(int(number), {"fg": TKCOLOR.red})


def show_widget():
    global root
    root = widget_root.main()
    sv_ttk.set_theme("dark")
    DownloadList(root).pack(anchor="center")
    tk.Frame(root, bg=TKCOLOR.dark_grey).pack(anchor="center", expand=True)
    root.mainloop()



import decimal
import re

import cloudscraper

from subsearch.data import __version__


def re_version(version: str, semantic: bool = False) -> str:
    try:
        pre_num = re.findall(f"\d*\.\d*\.\d*-\w*\.(\d*)", version)[0]
    except IndexError:
        pre_num = 0

    if semantic:
        # semantic expression https://regex101.com/r/2PNppl/1
        version_semantic = "".join(re.findall("(\d*\.\d*\.\d*-\w*\.\d*)|(\d*\.\d*\.\d*)", version)[0])
        return str(f"{version_semantic}")
    else:
        # float expression https://regex101.com/r/wt2fo6/1
        version_num = "".join(re.findall("(\d*)\.(\d*)\.(\d*)-\w*\.(\d*)|(\d*)\.(\d*)\.(\d*)", version)[0])
        return str(f"{version_num}.{pre_num}")


def scrape_github() -> str:
    """
    Scrape file  https://raw.githubusercontent.com/vagabondHustler/subsearch/main/src/subsearch/data/__version__.py

    Returns:
        str: returns the content of __version__.py
    """
    scraper = cloudscraper.create_scraper(browser={"browser": "chrome", "platform": "android", "desktop": False})
    url = "https://raw.githubusercontent.com/vagabondHustler/subsearch/main/src/subsearch/data/__version__.py"
    source = scraper.get(url)
    scontent = source.content
    file_content = str(scontent)
    return file_content


def get_current_version(semantic: bool = False) -> str:
    if semantic:
        return __version__
    return re_version(__version__)


def get_latest_version(semantic: bool = False) -> str:
    """
    Get the latest version number from subsearch github repo.
    As either a float or as semantic in str format

    Args:
        semantic (bool, optional): Decide which typ of string to return. Defaults to False.

    Returns:
        str:  xxx.x | x.x.x / x.x.x-type.x
    """
    if semantic:
        version_github = scrape_github()
        return re_version(version_github, True)

    version_github = scrape_github()
    return re_version(version_github)


def define_pre_release_worth(cver: str, csemantic: str, lver: str, lsemantic: str) -> tuple[float, float]:
    """
    Add int to float if rc, beta or alpha is found in semantic version

    rc = 3, beta = 2, alpha = 1

    x.x.x-beta.1 would equal float(2.1)

    Args:
        cver (str): current version as xxx.x
        csemantic (str): current version as semantic
        lver (str): latest version as xxx.x
        lsemantic (str): latest version as semantic

    Returns:
        tuple[float, float]:
    """
    # cver and lver needs to be a str otherwise it would return as 0.10.....something and not 0.1
    current_float = float(decimal.Decimal(cver) % 1)
    latest_float = float(decimal.Decimal(lver) % 1)
    semantic_versions = [csemantic, lsemantic]
    for num, item in enumerate(semantic_versions):
        if "-rc" in item:
            if num == 0:
                current_float += 3
            if num == 1:
                latest_float += 3
        elif "-beta" in item:
            if num == 0:
                current_float += 2
            if num == 1:
                latest_float += 2
        elif "-alpha" in item:
            if num == 0:
                current_float += 1
            if num == 1:
                latest_float += 1

    return current_float, latest_float


def is_new_version_avail() -> tuple[bool, bool]:
    """
    Check if there is a new version available over at the subsearch repo

    Returns:
        tuple[bool, bool]:
        stable release if True and False, pre-release if True and True
    """
    new_version = False
    pre_release = False

    curr_ver, curr_semantic = get_current_version(), get_current_version(semantic=True)
    ltst_ver, ltst_semantic = get_latest_version(), get_latest_version(semantic=True)
    curr_float, ltst_float = define_pre_release_worth(curr_ver, curr_semantic, ltst_ver, ltst_semantic)

    if curr_ver < ltst_ver:
        if curr_float < ltst_float:
            pre_release = True
        else:
            pre_release = False
        new_version = True

    return new_version, pre_release



import logging
import os
import sys
from datetime import datetime
from typing import Literal, Optional

from subsearch.data import __version__, __video_directory__
from subsearch.utils.string_parser import SearchParameters

CWD = os.getcwd()
if __video_directory__ is not None:
    CWD = __video_directory__


NOW = datetime.now()
DATE = NOW.strftime("%y%m%d")

logging.basicConfig(
    filename=f"{CWD}\\__subsearch__.log",
    filemode="w",
    level=logging.DEBUG,
    format="%(asctime)s - %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
)

# log and print message
def output(msg: str, to_terminal: bool = True) -> Optional[Literal[False]]:
    logging.info(msg)
    return print(msg) if to_terminal else False


def tprint(msg: str) -> None:
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    root.addHandler(handler)
    return logging.info(msg)


# log and print all the used parameters from video/directory-name
def parameters(param: SearchParameters, lang: str, lang_code2: str, hearing_impaired: str, pct: int) -> None:
    lang_code3 = lang[:3].lower()
    output(f"SubSearch - {__version__} ", False)
    output("[PARAMETERS]")
    output(f"Language: {lang}, {lang_code2}, {lang_code3}")
    output(f"Hearing impaired: {hearing_impaired}")
    output(f"Title: {param.title}")
    output(f"Year: {param.year}")
    output(f"Season: {param.season}, {param.season_ordinal}")
    output(f"Episode: {param.episode}, {param.episode_ordinal}")
    output(f"TV-Series: {param.show_bool}")
    output(f"Release: {param.release}")
    output(f"Group: {param.group}")
    output(f"Match threshold: {pct}%")
    output(f"File hash: {param.file_hash}")
    output(f"URL: {param.url_subscene}")
    output(f"URL: {param.url_opensubtitles}\n")



import re
from dataclasses import dataclass
from typing import Any, Literal, Optional, Union

from num2words import num2words


def find_year(string: str) -> Union[int, str]:
    """
    Parse string from start, until last instance of a year between .1000-2999. found, keep last instance of year
    https://regex101.com/r/r5TwxJ/1

    Args:
        string (str): Title.Of.The.Movie.YEAR.Source.Codec-GROUP

    Returns:
        str: YEAR
    """
    # from start of string until number between 1000-2999 ending with. found
    _year = re.findall("^.*\.([1-2][0-9]{3})\.", string)
    if len(_year) > 0:
        year = _year[0]
        return int(year)
    return "N/A"


def find_title_by_year(string: str) -> str:
    """
    Parse string from start, until last instance of a year between .1000-2999. found, keep everything before last instance of .year
    https://regex101.com/r/FKUpY0/1

    Args:
        string (str): Title.Of.The.Movie.YEAR.Source.Codec-GROUP

    Returns:
        str: Title.Of.The.Movie
    """
    _title = re.findall("^(.*)\.[1-2][0-9]{3}\.", string)
    if len(_title) > 0:
        title: str = _title[0]
        title = title.replace(".", " ")
        return title
    return "N/A"


def find_title_by_show(string: str) -> str:
    """
    Parse string from start, until last instance of .s00e00. found, keep everything before .season
    https://regex101.com/r/41OZE5/1

    Args:
        string (str): Title.Of.The.Show.s01e01.Source.Codec-GROUP

    Returns:
        str: Title.Of.The.Show
    """
    _title = re.findall("^(.*)\.[s]\d*[e]\d*\.", string)
    if len(_title) > 0:
        title: str = _title[0]
        title = title.replace(".", " ")
        return title
    return "N/A"


def find_season_episode(string: str) -> str:
    """
    Parse string from start, until last instance of .s00e00. found, keep .s00e00.
    https://regex101.com/r/8Nwlr6/1

    Args:
        string (str): Title.Of.The.Show.s01e01.Source.Codec-GROUP

    Returns:
        str: s01e01
    """
    _se = re.findall("\.([s]\d*[e]\d*)\.", string)
    if len(_se) > 0:
        se: str = _se[0]
        return se
    return "N/A"


NA = tuple[Literal["N/A"], Literal["N/A"], Literal["N/A"], Literal["N/A"], Literal[False]]


def find_ordinal(string: str, lang_code2: str) -> Union[tuple[str, str, str, str, bool], NA]:
    """
    Convert numbers into ordinal strings, 01 = First, 02 = Second...

    Args:
        string (str): s01e01
        lang_abbr (str): abbreviation of ordinal language

    Returns:
        str | int: _description_
    """
    if string == "N/A":
        season, season_ordinal, episode, episode_ordinal = "N/A", "N/A", "N/A", "N/A"
        show_bool = False

        return season, season_ordinal, episode, episode_ordinal, show_bool
    else:
        season, episode = string.replace("s", "").replace("e", " ").split(" ")
        season_ordinal = num2words(int(season), lang=lang_code2, to="ordinal")
        episode_ordinal = num2words(int(episode), lang=lang_code2, to="ordinal")
        show_bool = True
        return season, season_ordinal, episode, episode_ordinal, show_bool


def find_group(string: str) -> str:
    group = string.rsplit("-", 1)[-1]
    return group


@dataclass(frozen=True, order=True)
class SearchParameters:
    url_subscene: str
    url_opensubtitles: str
    title: str
    year: Union[int, str]
    season: str
    season_ordinal: str
    episode: str
    episode_ordinal: str
    show_bool: bool
    release: str
    group: str
    file_hash: Optional[str]


def get_parameters(filename: str, file_hash: Optional[str], language: str, lang_code2: str) -> SearchParameters:
    """
    Parse filename and get parameters for searching on subscene and opensubtitles
    Uses regex expressions to find the parameters

    Args:
        file_name (str): name of the file
        file_hash (str): hash of the file
        lang_abbr (str): language abbreviation for ordinal numbers

    Returns:
        SearchParameters: title, year, season, season_ordinal, episode, episode_ordinal, tv_series, release, group
    """
    filename = filename.lower()
    lang_code3 = language[:3].lower()

    year = find_year(filename)

    season_episode = find_season_episode(filename)
    season, season_ordinal, episode, episode_ordinal, show_bool = find_ordinal(season_episode, lang_code2)

    if year != "N/A":
        title = find_title_by_year(filename)
    elif show_bool and year == "N/A":
        title = find_title_by_show(filename)
        title = f"{title} - {season_ordinal} season"
    else:
        title = filename.rsplit("-", 1)[0]

    group = find_group(filename)
    subscene = "https://subscene.com/subtitles/searchbytitle?query="
    opensubtitles = f"https://www.opensubtitles.org/en/search/sublanguageid-{lang_code3}/moviename-"
    url_subscene = f"{subscene}{title}".replace(" ", "%20")
    url_opensubtitles = f"{opensubtitles}{file_hash}"

    parameters = SearchParameters(
        url_subscene,
        url_opensubtitles,
        title,
        year,
        season,
        season_ordinal,
        episode,
        episode_ordinal,
        show_bool,
        filename,
        group,
        file_hash,
    )
    return parameters


def get_pct_value(from_pc: Any, from_browser: Any) -> int:
    """
    Parse two lists and return match in %

    Args:
        from_pc (str | int): list from pc
        from_browser (str | int): list from browser
    Returns:
        int: value in percentage of how many words match
    """
    max_percentage = 100
    pc_list: list[Any] = mk_lst(from_pc)
    browser_list: list[Any] = mk_lst(from_browser)
    not_matching = list(set(pc_list) - set(browser_list))
    not_matching_value = len(not_matching)
    number_of_items = len(pc_list)
    percentage_to_remove = round(not_matching_value / number_of_items * max_percentage)
    pct = round(max_percentage - percentage_to_remove)

    return pct


# create list from string
def mk_lst(release: Any) -> list[Any]:
    new: list[Any] = []
    qualities = ["720p", "1080p", "1440p", "2160p"]
    temp = release.split(".")

    for item in temp:
        if item not in qualities:
            new.append(item.lower())
    return new
